# Copyright 2023 UW-IT, University of Washington
# SPDX-License-Identifier: Apache-2.0


import json
from copy import copy


class AbstractBase():

    def __getattr__(self, name):
        try:
            return self.__dict__[name]
        except KeyError:
            raise AttributeError(f"Attribute '{name}' does not exist for "
                                 f"'{type(self).__name__}' instance.")

    def __setattr__(self, name, value):
        self.__dict__[name] = value

    def to_dict(self):

        def format(value):
            if isinstance(value, AbstractBase):
                return value.to_dict()
            elif isinstance(value, list):
                return [format(item) for item in value]
            else:
                return value

        data = copy(self.__dict__)
        for key, value in data.items():
            data[key] = format(value)
        return data

    def to_json(self):
        return json.dumps(self.to_dict(),
                          sort_keys=True,
                          indent=2,
                          default=str)

    def from_dict(self, data, obj=None):
        if obj is None:
            obj = self
        for key, value in data.items():
            if isinstance(value, dict):
                if key == "employee":
                    new_obj = Employee()
                    obj.employee = new_obj.from_dict(value, obj=new_obj)
                elif key == "adviser":
                    new_obj = Adviser()
                    obj.adviser = new_obj.from_dict(value, obj=new_obj)
                elif key == "student":
                    new_obj = Student()
                    obj.student = new_obj.from_dict(value, obj=new_obj)
                elif key == "tran_term":
                    new_obj = Term()
                    obj.tran_term = new_obj.from_dict(value, obj=new_obj)
                elif key == "leave_ends_term":
                    new_obj = Term()
                    obj.leave_ends_term = new_obj.from_dict(value, obj=new_obj)
            elif isinstance(value, list):
                if key == "majors" or key == "pending_majors":
                    obj_cls = Major
                elif key == "advisers":
                    obj_cls = Adviser
                elif key == "transcripts":
                    obj_cls = Transcript
                elif key == "transfers":
                    obj_cls = Transfer
                elif key == "sports":
                    obj_cls = Sport
                elif key == "holds":
                    obj_cls = Hold
                items = []
                for list_value in value:
                    if isinstance(list_value, dict):
                        new_obj = obj_cls()
                        items.append(
                            new_obj.from_dict(list_value, obj=new_obj))
                    else:
                        items.append(list_value)
                obj.__setattr__(key, items)
            else:
                obj.__setattr__(key, value)
        return obj


class Person(AbstractBase):
    pass


class Student(AbstractBase):
    ETHNIC_GROUP_DESCRIPTIONS = {
        "1": "African American",
        "2": "American Indian",
        "3": "White",
        "4": "Hispanic/Latino",
        "5": "Asian American",
        "6": "Hawaiian/Pacific Islander",
        "7": "Not Indicated",
        "99": "International",
    }


class Employee(AbstractBase):
    pass


class Transcript(AbstractBase):
    pass


class Transfer(AbstractBase):
    pass


class Hold(AbstractBase):
    TYPE_DESCRIPTIONS = {
        1: "REGISTRATION HOLD",
        2: "TRANSCRIPT HOLD",
        3: "REGISTRATION AND TRANSCRIPT HOLD"
    }
    OFFICE_DESCRIPTIONS = {
        "EOP": "EDUCATIONAL OPPTNY PROGR"
    }


class Degree(AbstractBase):
    GRAD_HONOR_DESCRIPTIONS = {
        1: "SUMMA CUM LAUDE",
        2: "MAGNA CUM LAUDE",
        3: "CUM LAUDE"
    }


class Major(AbstractBase):
    CAMPUS_NAMES = {
        0: "Seattle",
        1: "Bothell",
        2: "Tacoma"
    }
    COLLEGE_FULL_NAMES = {
        "A": "Interdisciplinary Undergraduate Programs",
        "B": "College of Built Environments",
        "C": "College of Arts & Sciences",
        "D": "College of the Environment",
        "E": "Foster School of Business",
        "H": "College of Education",
        "J": "College of Engineering",
        "J2": "School of Computer Science & Engineering",
        "K": "College of Ocean & Fishery Sciences",
        "L": "College of Forest Resources",
        "M": "School of Public Health",
        "N": "School of Nursing",
        "O": "Interschool or Intercollege Programs",
        "P": "School of Pharmacy",
        "Q": "Evans School of Public Affairs",
        "R": "Interdisciplinary Graduate Programs",
        "S": "The Information School",
        "T": "School of Social Work",
        "U": "School of Dentistry",
        "V": "UW Bothell",
        "X": "School of Law",
        "Y": "School of Medicine",
        "Z": "UW Tacoma"
    }


class Sport(AbstractBase):
    pass


class Adviser(AbstractBase):
    pass


class Term(AbstractBase):
    TERM_NAMES = {1: "Winter", 2: "Spring", 3: "Summer", 4: "Autumn"}
